@@ -14,7 +14,7 @@
 /**
  * static_assert() with an optional second argument.
  */
-#if __STDC_VERSION__ >= 202311L
+#if __STDC_VERSION__ >= C23
 #  define bfs_static_assert static_assert
 #else
 #  define bfs_static_assert(...) bfs_static_assert_(__VA_ARGS__, #__VA_ARGS__, )
@@ -35,7 +35,7 @@ struct bfs_loc {
 /**
  * Get the current source code location.
  */
-#if __STDC_VERSION__ >= 202311L
+#if __STDC_VERSION__ >= C23
 #  define bfs_location() (&(static const struct bfs_loc)BFS_LOC_INIT)
 #else
 #  define bfs_location() (&(const struct bfs_loc)BFS_LOC_INIT)
@@ -44,7 +44,7 @@ struct bfs_loc {
 /**
  * Print a message to standard error and abort.
  */
-BFS_FORMATTER(2, 3)
+attr(cold, printf(2, 3))
 noreturn void bfs_abortf(const struct bfs_loc *loc, const char *format, ...);
 
 /**
@@ -116,12 +116,13 @@ const char *debug_flag_name(enum debug_flags flag);
 /**
  * Like perror(), but decorated like bfs_error().
  */
+attr(cold)
 void bfs_perror(const struct bfs_ctx *ctx, const char *str);
 
 /**
  * Shorthand for printing error messages.
  */
-BFS_FORMATTER(2, 3)
+attr(cold, printf(2, 3))
 void bfs_error(const struct bfs_ctx *ctx, const char *format, ...);
 
 /**
@@ -129,7 +130,7 @@ void bfs_error(const struct bfs_ctx *ctx, const char *format, ...);
  *
  * @return Whether a warning was printed.
  */
-BFS_FORMATTER(2, 3)
+attr(cold, printf(2, 3))
 bool bfs_warning(const struct bfs_ctx *ctx, const char *format, ...);
 
 /**
@@ -137,60 +138,67 @@ bool bfs_warning(const struct bfs_ctx *ctx, const char *format, ...);
  *
  * @return Whether a debug message was printed.
  */
-BFS_FORMATTER(3, 4)
+attr(cold, printf(3, 4))
 bool bfs_debug(const struct bfs_ctx *ctx, enum debug_flags flag, const char *format, ...);
 
 /**
  * bfs_error() variant that takes a va_list.
  */
-BFS_FORMATTER(2, 0)
+attr(cold, printf(2, 0))
 void bfs_verror(const struct bfs_ctx *ctx, const char *format, va_list args);
 
 /**
  * bfs_warning() variant that takes a va_list.
  */
-BFS_FORMATTER(2, 0)
+attr(cold, printf(2, 0))
 bool bfs_vwarning(const struct bfs_ctx *ctx, const char *format, va_list args);
 
 /**
  * bfs_debug() variant that takes a va_list.
  */
-BFS_FORMATTER(3, 0)
+attr(cold, printf(3, 0))
 bool bfs_vdebug(const struct bfs_ctx *ctx, enum debug_flags flag, const char *format, va_list args);
 
 /**
  * Print the error message prefix.
  */
+attr(cold)
 void bfs_error_prefix(const struct bfs_ctx *ctx);
 
 /**
  * Print the warning message prefix.
  */
+attr(cold)
 bool bfs_warning_prefix(const struct bfs_ctx *ctx);
 
 /**
  * Print the debug message prefix.
  */
+attr(cold)
 bool bfs_debug_prefix(const struct bfs_ctx *ctx, enum debug_flags flag);
 
 /**
  * Highlight parts of the command line in an error message.
  */
+attr(cold)
 void bfs_argv_error(const struct bfs_ctx *ctx, const bool args[]);
 
 /**
  * Highlight parts of an expression in an error message.
  */
+attr(cold)
 void bfs_expr_error(const struct bfs_ctx *ctx, const struct bfs_expr *expr);
 
 /**
  * Highlight parts of the command line in a warning message.
  */
+attr(cold)
 bool bfs_argv_warning(const struct bfs_ctx *ctx, const bool args[]);
 
 /**
  * Highlight parts of an expression in a warning message.
  */
+attr(cold)
 bool bfs_expr_warning(const struct bfs_ctx *ctx, const struct bfs_expr *expr);
 
 #endif // BFS_DIAG_H
