@@ -1,6 +1,127 @@
 // Copyright © Tavian Barnes <tavianator@tavianator.com>
 // SPDX-License-Identifier: 0BSD
 
+/**
+ * An asynchronous I/O queue implementation.
+ *
+ * struct ioq is composed of two separate queues:
+ *
+ *     struct ioqq *pending; // Pending I/O requests
+ *     struct ioqq *ready;   // Ready I/O responses
+ *
+ * Worker threads pop requests from `pending`, execute them, and push them back
+ * to the `ready` queue.  The main thread pushes requests to `pending` and pops
+ * them from `ready`.
+ *
+ * struct ioqq is a blocking MPMC queue (though it could be SPMC/MPSC for
+ * pending/ready respectively).  It is implemented as a circular buffer:
+ *
+ *     size_t mask;            // (1 << N) - 1
+ *     [padding]
+ *     size_t head;            // Writer index
+ *     [padding]
+ *     size_t tail;            // Reader index
+ *     [padding]
+ *     ioq_slot slots[1 << N]; // Queue contents
+ *
+ * Pushes are implemented with an unconditional
+ *
+ *     fetch_add(&ioqq->head, IOQ_STRIDE)
+ *
+ * which scales better on many architectures than compare-and-swap (see [1] for
+ * details).  Pops are implemented similarly.  We add IOQ_STRIDE rather than 1
+ * so that successive queue elements are on different cache lines, but the
+ * exposition below uses 1 for simplicity.
+ *
+ * Since the fetch-and-adds are unconditional, non-blocking readers can get
+ * ahead of writers:
+ *
+ *     Reader              Writer
+ *     ────────────────    ──────────────────────
+ *     head:     0 → 1
+ *     slots[0]: empty
+ *                         tail:     0     → 1
+ *                         slots[0]: empty → full
+ *     head:     1 → 2
+ *     slots[1]: empty!
+ *
+ * To avoid this, non-blocking reads (ioqq_pop(ioqq, false)) must mark the slots
+ * somehow so that writers can skip them:
+ *
+ *     Reader                     Writer
+ *     ───────────────────────    ───────────────────────
+ *     head:         0 → 1
+ *     slots[0]: empty → skip
+ *                                tail:         0 → 1
+ *                                slots[0]:  skip → empty
+ *                                tail:         1 → 2
+ *                                slots[1]: empty → full
+ *     head:         1 → 2
+ *     slots[1]:  full → empty
+ *
+ * As well, a reader might "lap" a writer (or another reader), so slots need to
+ * count how many times they should be skipped:
+ *
+ *     Reader                        Writer
+ *     ──────────────────────────    ─────────────────────────
+ *     head:          0 → 1
+ *     slots[0]:  empty → skip(1)
+ *     head:          1 → 2
+ *     slots[1]:  empty → skip(1)
+ *     ...
+ *     head:          M → 0
+ *     slots[M]:  empty → skip(1)
+ *     head:          0 → 1
+ *     slots[0]: skip(1 → 2)
+ *                                   tail:           0 → 1
+ *                                   slots[0]:  skip(2 → 1)
+ *                                   tail:           1 → 2
+ *                                   slots[1]: skip(1) → empty
+ *                                   ...
+ *                                   tail:           M → 0
+ *                                   slots[M]: skip(1) → empty
+ *                                   tail:           0 → 1
+ *                                   slots[0]: skip(1) → empty
+ *                                   tail:           1 → 2
+ *                                   slots[1]:   empty → full
+ *     head:          1 → 2
+ *     slots[1]:   full → empty
+ *
+ * As described in [1], this approach is susceptible to livelock if readers stay
+ * ahead of writers.  This is okay for us because we don't retry failed non-
+ * blocking reads.
+ *
+ * The slot representation uses tag bits to hold either a pointer or skip(N):
+ *
+ *     IOQ_SKIP (highest bit)    IOQ_BLOCKED (lowest bit)
+ *        ↓                         ↓
+ *        0 0 0       ...       0 0 0
+ *          └──────────┬──────────┘
+ *                     │
+ *                value bits
+ *
+ * If IOQ_SKIP is unset, the value bits hold a pointer (or zero/NULL for empty).
+ * If IOQ_SKIP is set, the value bits hold a negative skip count.  Writers can
+ * reduce the skip count by adding 1 to the value bits, and when the count hits
+ * zero, the carry will automatically clear IOQ_SKIP:
+ *
+ *     IOQ_SKIP                  IOQ_BLOCKED
+ *        ↓                         ↓
+ *        1 1 1       ...       1 0 0    skip(2)
+ *        1 1 1       ...       1 1 0    skip(1)
+ *        0 0 0       ...       0 0 0    empty
+ *
+ * The IOQ_BLOCKED flag is used to track sleeping waiters, futex-style.  To wait
+ * for a slot to change, waiters call ioq_slot_wait() which sets IOQ_BLOCKED and
+ * goes to sleep.  Whenever a slot is updated, if the old value had IOQ_BLOCKED
+ * set, ioq_slot_wake() must be called to wake up that waiter.
+ *
+ * Blocking/waking uses a pool of monitors (mutex, condition variable pairs).
+ * Slots are assigned round-robin to a monitor from the pool.
+ *
+ * [1]: https://arxiv.org/abs/2201.02179
+ */
+
 #include "ioq.h"
 #include "alloc.h"
 #include "atomic.h"
@@ -10,6 +131,7 @@
 #include "diag.h"
 #include "dir.h"
 #include "sanity.h"
+#include "stat.h"
 #include "thread.h"
 #include <assert.h>
 #include <errno.h>
@@ -51,24 +173,15 @@ static void ioq_monitor_destroy(struct ioq_monitor *monitor) {
 /** A single entry in a command queue. */
 typedef atomic uintptr_t ioq_slot;
 
-/** Slot flag bit to indicate waiters. */
+/** Someone might be waiting on this slot. */
 #define IOQ_BLOCKED ((uintptr_t)1)
-bfs_static_assert(alignof(struct ioq_ent) > 1);
+/** The next push(es) should skip this slot. */
+#define IOQ_SKIP ((uintptr_t)1 << (UINTPTR_WIDTH - 1))
+/** Amount to add for an additional skip. */
+#define IOQ_SKIP_ONE (~IOQ_BLOCKED)
 
-/** Check if a slot has waiters. */
-static bool ioq_slot_blocked(uintptr_t value) {
-	return value & IOQ_BLOCKED;
-}
-
-/** Extract the pointer from a slot. */
-static struct ioq_ent *ioq_slot_ptr(uintptr_t value) {
-	return (struct ioq_ent *)(value & ~IOQ_BLOCKED);
-}
-
-/** Check if a slot is empty. */
-static bool ioq_slot_empty(uintptr_t value) {
-	return !ioq_slot_ptr(value);
-}
+// Need room for two flag bits
+bfs_static_assert(alignof(struct ioq_ent) > 2);
 
 /**
  * An MPMC queue of I/O commands.
@@ -104,6 +217,10 @@ bfs_static_assert(IOQ_STRIDE % 2 == 1);
 
 /** Destroy an I/O command queue. */
 static void ioqq_destroy(struct ioqq *ioqq) {
+	if (!ioqq) {
+		return;
+	}
+
 	for (size_t i = 0; i < ioqq->monitor_mask + 1; ++i) {
 		ioq_monitor_destroy(&ioqq->monitors[i]);
 	}
@@ -157,6 +274,7 @@ static struct ioq_monitor *ioq_slot_monitor(struct ioqq *ioqq, ioq_slot *slot) {
 }
 
 /** Atomically wait for a slot to change. */
+attr(noinline)
 static uintptr_t ioq_slot_wait(struct ioqq *ioqq, ioq_slot *slot, uintptr_t value) {
 	struct ioq_monitor *monitor = ioq_slot_monitor(ioqq, slot);
 	mutex_lock(&monitor->mutex);
@@ -186,6 +304,7 @@ done:
 }
 
 /** Wake up any threads waiting on a slot. */
+attr(noinline)
 static void ioq_slot_wake(struct ioqq *ioqq, ioq_slot *slot) {
 	struct ioq_monitor *monitor = ioq_slot_monitor(ioqq, slot);
 
@@ -205,80 +324,85 @@ static void ioq_slot_wake(struct ioqq *ioqq, ioq_slot *slot) {
 	cond_broadcast(&monitor->cond);
 }
 
-/** Get the next slot for writing. */
-static ioq_slot *ioqq_write(struct ioqq *ioqq) {
-	size_t i = fetch_add(&ioqq->head, IOQ_STRIDE, relaxed);
-	return &ioqq->slots[i & ioqq->slot_mask];
-}
-
 /** Push an entry into a slot. */
-static void ioq_slot_push(struct ioqq *ioqq, ioq_slot *slot, struct ioq_ent *ent) {
-	uintptr_t addr = (uintptr_t)ent;
-	bfs_assert(!ioq_slot_blocked(addr));
-
+static bool ioq_slot_push(struct ioqq *ioqq, ioq_slot *slot, struct ioq_ent *ent) {
 	uintptr_t prev = load(slot, relaxed);
-	do {
-		while (!ioq_slot_empty(prev)) {
+	while (true) {
+		uintptr_t next;
+		if (prev & IOQ_SKIP) {
+			// skip(1) → empty
+			// skip(n) → skip(n - 1)
+			next = (prev - IOQ_SKIP_ONE) & ~IOQ_BLOCKED;
+		} else if (prev > IOQ_BLOCKED) {
+			// full(ptr) → wait
 			prev = ioq_slot_wait(ioqq, slot, prev);
+			continue;
+		} else {
+			// empty → full(ptr)
+			next = (uintptr_t)ent >> 1;
 		}
-	} while (!compare_exchange_weak(slot, &prev, addr, release, relaxed));
 
-	if (ioq_slot_blocked(prev)) {
+		if (compare_exchange_weak(slot, &prev, next, release, relaxed)) {
+			break;
+		}
+	}
+
+	if (prev & IOQ_BLOCKED) {
 		ioq_slot_wake(ioqq, slot);
 	}
+
+	return !(prev & IOQ_SKIP);
 }
 
 /** Push an entry onto the queue. */
 static void ioqq_push(struct ioqq *ioqq, struct ioq_ent *ent) {
-	ioq_slot *slot = ioqq_write(ioqq);
-	ioq_slot_push(ioqq, slot, ent);
-}
-
-/** Get the next slot for reading. */
-static ioq_slot *ioqq_read(struct ioqq *ioqq) {
-	size_t i = fetch_add(&ioqq->tail, IOQ_STRIDE, relaxed);
-	return &ioqq->slots[i & ioqq->slot_mask];
+	while (true) {
+		size_t i = fetch_add(&ioqq->head, IOQ_STRIDE, relaxed);
+		ioq_slot *slot = &ioqq->slots[i & ioqq->slot_mask];
+		if (ioq_slot_push(ioqq, slot, ent)) {
+			break;
+		}
+	}
 }
 
 /** (Try to) pop an entry from a slot. */
 static struct ioq_ent *ioq_slot_pop(struct ioqq *ioqq, ioq_slot *slot, bool block) {
 	uintptr_t prev = load(slot, relaxed);
-	do {
-		while (ioq_slot_empty(prev)) {
-			if (block) {
-				prev = ioq_slot_wait(ioqq, slot, prev);
-			} else {
-				return NULL;
-			}
-		}
-	} while (!compare_exchange_weak(slot, &prev, 0, acquire, relaxed));
+	while (true) {
+		// empty     → skip(1)
+		// skip(n)   → skip(n + 1)
+		// full(ptr) → full(ptr - 1)
+		uintptr_t next = prev + IOQ_SKIP_ONE;
+		// skip(n)   → ~IOQ_BLOCKED
+		// full(ptr) → 0
+		next &= (next & IOQ_SKIP) ? ~IOQ_BLOCKED : 0;
 
-	if (ioq_slot_blocked(prev)) {
+		if (block && next) {
+			prev = ioq_slot_wait(ioqq, slot, prev);
+			continue;
+		}
+
+		if (compare_exchange_weak(slot, &prev, next, acquire, relaxed)) {
+			break;
+		}
+	}
+
+	if (prev & IOQ_BLOCKED) {
 		ioq_slot_wake(ioqq, slot);
 	}
 
-	return ioq_slot_ptr(prev);
+	// empty     → 0
+	// skip(n)   → 0
+	// full(ptr) → ptr
+	prev &= (prev & IOQ_SKIP) ? 0 : ~IOQ_BLOCKED;
+	return (struct ioq_ent *)(prev << 1);
 }
 
 /** Pop an entry from the queue. */
-static struct ioq_ent *ioqq_pop(struct ioqq *ioqq) {
-	ioq_slot *slot = ioqq_read(ioqq);
-	return ioq_slot_pop(ioqq, slot, true);
-}
-
-/** Pop an entry from the queue if one is available. */
-static struct ioq_ent *ioqq_trypop(struct ioqq *ioqq) {
-	size_t i = load(&ioqq->tail, relaxed);
+static struct ioq_ent *ioqq_pop(struct ioqq *ioqq, bool block) {
+	size_t i = fetch_add(&ioqq->tail, IOQ_STRIDE, relaxed);
 	ioq_slot *slot = &ioqq->slots[i & ioqq->slot_mask];
-
-	struct ioq_ent *ret = ioq_slot_pop(ioqq, slot, false);
-	if (ret) {
-		size_t j = exchange(&ioqq->tail, i + IOQ_STRIDE, relaxed);
-		bfs_assert(j == i, "Detected multiple consumers");
-		(void)j;
-	}
-
-	return ret;
+	return ioq_slot_pop(ioqq, slot, block);
 }
 
 /** Sentinel stop command. */
@@ -309,6 +433,10 @@ struct ioq {
 
 	/** ioq_ent arena. */
 	struct arena ents;
+#if BFS_USE_LIBURING && BFS_USE_STATX
+	/** struct statx arena. */
+	struct arena xbufs;
+#endif
 
 	/** Pending I/O requests. */
 	struct ioqq *pending;
@@ -332,54 +460,56 @@ static bool ioq_check_cancel(struct ioq *ioq, struct ioq_ent *ent) {
 		return false;
 	}
 
-	ent->ret = -1;
-	ent->error = EINTR;
+	ent->result = -EINTR;
 	ioqq_push(ioq->ready, ent);
 	return true;
 }
 
-/** Handle a single request synchronously. */
-static void ioq_handle(struct ioq *ioq, struct ioq_ent *ent) {
-	int ret;
-
+/** Dispatch a single request synchronously. */
+static void ioq_dispatch_sync(struct ioq *ioq, struct ioq_ent *ent) {
 	switch (ent->op) {
-	case IOQ_CLOSE:
-		ret = xclose(ent->close.fd);
-		break;
+		case IOQ_CLOSE:
+			ent->result = try(xclose(ent->close.fd));
+			return;
 
-	case IOQ_OPENDIR:
-		ret = bfs_opendir(ent->opendir.dir, ent->opendir.dfd, ent->opendir.path);
-		if (ret == 0) {
-			bfs_polldir(ent->opendir.dir);
+		case IOQ_OPENDIR: {
+			struct ioq_opendir *args = &ent->opendir;
+			ent->result = try(bfs_opendir(args->dir, args->dfd, args->path, args->flags));
+			if (ent->result >= 0) {
+				bfs_polldir(args->dir);
+			}
+			return;
 		}
-		break;
 
-	case IOQ_CLOSEDIR:
-		ret = bfs_closedir(ent->closedir.dir);
-		break;
+		case IOQ_CLOSEDIR:
+			ent->result = try(bfs_closedir(ent->closedir.dir));
+			return;
 
-	default:
-		bfs_bug("Unknown ioq_op %d", (int)ent->op);
-		ret = -1;
-		errno = ENOSYS;
-		break;
+		case IOQ_STAT: {
+			struct ioq_stat *args = &ent->stat;
+			ent->result = try(bfs_stat(args->dfd, args->path, args->flags, args->buf));
+			return;
+		}
 	}
 
-	ent->ret = ret;
-	ent->error = ret == 0 ? 0 : errno;
+	bfs_bug("Unknown ioq_op %d", (int)ent->op);
+	ent->result = -ENOSYS;
+}
 
+/** Complete a single request synchronously. */
+static void ioq_complete(struct ioq *ioq, struct ioq_ent *ent) {
+	ioq_dispatch_sync(ioq, ent);
 	ioqq_push(ioq->ready, ent);
 }
 
 #if BFS_USE_LIBURING
+
 /** io_uring worker state. */
 struct ioq_ring_state {
 	/** The I/O queue. */
 	struct ioq *ioq;
 	/** The io_uring. */
 	struct io_uring *ring;
-	/** The current ioq->pending slot. */
-	ioq_slot *slot;
 	/** Number of prepped, unsubmitted SQEs. */
 	size_t prepped;
 	/** Number of submitted, unreaped SQEs. */
@@ -394,20 +524,9 @@ static struct ioq_ent *ioq_ring_pop(struct ioq_ring_state *state) {
 		return NULL;
 	}
 
-	// Advance to the next slot if necessary
-	struct ioq *ioq = state->ioq;
-	if (!state->slot) {
-		state->slot = ioqq_read(ioq->pending);
-	}
-
 	// Block if we have nothing else to do
 	bool block = !state->prepped && !state->submitted;
-	struct ioq_ent *ret = ioq_slot_pop(ioq->pending, state->slot, block);
-
-	if (ret) {
-		// Got an entry, move to the next slot next time
-		state->slot = NULL;
-	}
+	struct ioq_ent *ret = ioqq_pop(state->ioq->pending, block);
 
 	if (ret == &IOQ_STOP) {
 		state->stop = true;
@@ -417,35 +536,65 @@ static struct ioq_ent *ioq_ring_pop(struct ioq_ring_state *state) {
 	return ret;
 }
 
-/** Prep a single SQE. */
-static void ioq_prep_sqe(struct io_uring_sqe *sqe, struct ioq_ent *ent) {
+/** Dispatch a single request asynchronously. */
+static struct io_uring_sqe *ioq_dispatch_async(struct io_uring *ring, struct ioq_ent *ent) {
+	struct io_uring_sqe *sqe = NULL;
+
 	switch (ent->op) {
-	case IOQ_CLOSE:
-		io_uring_prep_close(sqe, ent->close.fd);
-		break;
+		case IOQ_CLOSE:
+			sqe = io_uring_get_sqe(ring);
+			io_uring_prep_close(sqe, ent->close.fd);
+			return sqe;
 
-	case IOQ_OPENDIR:
-		io_uring_prep_openat(sqe, ent->opendir.dfd, ent->opendir.path, O_RDONLY | O_CLOEXEC | O_DIRECTORY, 0);
-		break;
+		case IOQ_OPENDIR: {
+			sqe = io_uring_get_sqe(ring);
+			struct ioq_opendir *args = &ent->opendir;
+			int flags = O_RDONLY | O_CLOEXEC | O_DIRECTORY;
+			io_uring_prep_openat(sqe, args->dfd, args->path, flags, 0);
+			return sqe;
+		}
 
+		case IOQ_CLOSEDIR:
 #if BFS_USE_UNWRAPDIR
-	case IOQ_CLOSEDIR:
-		io_uring_prep_close(sqe, bfs_unwrapdir(ent->closedir.dir));
-		break;
+			sqe = io_uring_get_sqe(ring);
+			io_uring_prep_close(sqe, bfs_unwrapdir(ent->closedir.dir));
 #endif
+			return sqe;
 
-	default:
-		bfs_bug("Unknown ioq_op %d", (int)ent->op);
-		io_uring_prep_nop(sqe);
-		break;
+		case IOQ_STAT: {
+#if BFS_USE_STATX
+			sqe = io_uring_get_sqe(ring);
+			struct ioq_stat *args = &ent->stat;
+			int flags = bfs_statx_flags(args->flags);
+			unsigned int mask = STATX_BASIC_STATS | STATX_BTIME;
+			io_uring_prep_statx(sqe, args->dfd, args->path, flags, mask, args->xbuf);
+#endif
+			return sqe;
+		}
 	}
 
-	io_uring_sqe_set_data(sqe, ent);
+	bfs_bug("Unknown ioq_op %d", (int)ent->op);
+	return NULL;
+}
+
+/** Prep a single SQE. */
+static void ioq_prep_sqe(struct ioq_ring_state *state, struct ioq_ent *ent) {
+	struct ioq *ioq = state->ioq;
+	if (ioq_check_cancel(ioq, ent)) {
+		return;
+	}
+
+	struct io_uring_sqe *sqe = ioq_dispatch_async(state->ring, ent);
+	if (sqe) {
+		io_uring_sqe_set_data(sqe, ent);
+		++state->prepped;
+	} else {
+		ioq_complete(ioq, ent);
+	}
 }
 
 /** Prep a batch of SQEs. */
 static bool ioq_ring_prep(struct ioq_ring_state *state) {
-	struct ioq *ioq = state->ioq;
 	struct io_uring *ring = state->ring;
 
 	while (io_uring_sq_space_left(ring)) {
@@ -454,30 +603,64 @@ static bool ioq_ring_prep(struct ioq_ring_state *state) {
 			break;
 		}
 
-		if (ioq_check_cancel(ioq, ent)) {
-			continue;
-		}
-
-#if !BFS_USE_UNWRAPDIR
-		if (ent->op == IOQ_CLOSEDIR) {
-			ioq_handle(ioq, ent);
-			continue;
-		}
-#endif
-
-		struct io_uring_sqe *sqe = io_uring_get_sqe(ring);
-		ioq_prep_sqe(sqe, ent);
-		++state->prepped;
+		ioq_prep_sqe(state, ent);
 	}
 
 	return state->prepped || state->submitted;
 }
 
-/** Reap a batch of SQEs. */
-static void ioq_ring_reap(struct ioq_ring_state *state) {
+/** Reap a single CQE. */
+static void ioq_reap_cqe(struct ioq_ring_state *state, struct io_uring_cqe *cqe) {
 	struct ioq *ioq = state->ioq;
 	struct io_uring *ring = state->ring;
 
+	struct ioq_ent *ent = io_uring_cqe_get_data(cqe);
+	ent->result = cqe->res;
+	io_uring_cqe_seen(ring, cqe);
+	--state->submitted;
+
+	if (ent->result < 0) {
+		goto push;
+	}
+
+	switch (ent->op) {
+		case IOQ_OPENDIR: {
+			int fd = ent->result;
+			if (ioq_check_cancel(ioq, ent)) {
+				xclose(fd);
+				return;
+			}
+
+			struct ioq_opendir *args = &ent->opendir;
+			ent->result = try(bfs_opendir(args->dir, fd, NULL, args->flags));
+			if (ent->result >= 0) {
+				// TODO: io_uring_prep_getdents()
+				bfs_polldir(args->dir);
+			}
+
+			break;
+		}
+
+#if BFS_USE_STATX
+		case IOQ_STAT: {
+			struct ioq_stat *args = &ent->stat;
+			ent->result = try(bfs_statx_convert(args->buf, args->xbuf));
+			break;
+		}
+#endif
+
+		default:
+			break;
+	}
+
+push:
+	ioqq_push(ioq->ready, ent);
+}
+
+/** Reap a batch of CQEs. */
+static void ioq_ring_reap(struct ioq_ring_state *state) {
+	struct io_uring *ring = state->ring;
+
 	while (state->prepped) {
 		int ret = io_uring_submit_and_wait(ring, 1);
 		if (ret > 0) {
@@ -492,29 +675,7 @@ static void ioq_ring_reap(struct ioq_ring_state *state) {
 			continue;
 		}
 
-		struct ioq_ent *ent = io_uring_cqe_get_data(cqe);
-		ent->ret = cqe->res >= 0 ? cqe->res : -1;
-		ent->error = cqe->res < 0 ? -cqe->res : 0;
-		io_uring_cqe_seen(ring, cqe);
-		--state->submitted;
-
-		if (ent->op == IOQ_OPENDIR && ent->ret >= 0) {
-			int fd = ent->ret;
-			if (ioq_check_cancel(ioq, ent)) {
-				xclose(fd);
-				continue;
-			}
-
-			ent->ret = bfs_opendir(ent->opendir.dir, fd, NULL);
-			if (ent->ret == 0) {
-				// TODO: io_uring_prep_getdents()
-				bfs_polldir(ent->opendir.dir);
-			} else {
-				ent->error = errno;
-			}
-		}
-
-		ioqq_push(ioq->ready, ent);
+		ioq_reap_cqe(state, cqe);
 	}
 }
 
@@ -529,20 +690,21 @@ static void ioq_ring_work(struct ioq_thread *thread) {
 		ioq_ring_reap(&state);
 	}
 }
-#endif
+
+#endif // BFS_USE_LIBURING
 
 /** Synchronous syscall loop. */
 static void ioq_sync_work(struct ioq_thread *thread) {
 	struct ioq *ioq = thread->parent;
 
 	while (true) {
-		struct ioq_ent *ent = ioqq_pop(ioq->pending);
+		struct ioq_ent *ent = ioqq_pop(ioq->pending, true);
 		if (ent == &IOQ_STOP) {
 			break;
 		}
 
 		if (!ioq_check_cancel(ioq, ent)) {
-			ioq_handle(ioq, ent);
+			ioq_complete(ioq, ent);
 		}
 	}
 }
@@ -569,8 +731,13 @@ struct ioq *ioq_create(size_t depth, size_t nthreads) {
 	}
 
 	ioq->depth = depth;
+
 	ARENA_INIT(&ioq->ents, struct ioq_ent);
 
+#if BFS_USE_LIBURING && BFS_USE_STATX
+	ARENA_INIT(&ioq->xbufs, struct statx);
+#endif
+
 	ioq->pending = ioqq_create(depth);
 	if (!ioq->pending) {
 		goto fail;
@@ -659,7 +826,7 @@ int ioq_close(struct ioq *ioq, int fd, void *ptr) {
 	return 0;
 }
 
-int ioq_opendir(struct ioq *ioq, struct bfs_dir *dir, int dfd, const char *path, void *ptr) {
+int ioq_opendir(struct ioq *ioq, struct bfs_dir *dir, int dfd, const char *path, enum bfs_dir_flags flags, void *ptr) {
 	struct ioq_ent *ent = ioq_request(ioq, IOQ_OPENDIR, ptr);
 	if (!ent) {
 		return -1;
@@ -669,6 +836,7 @@ int ioq_opendir(struct ioq *ioq, struct bfs_dir *dir, int dfd, const char *path,
 	args->dir = dir;
 	args->dfd = dfd;
 	args->path = path;
+	args->flags = flags;
 
 	ioqq_push(ioq->pending, ent);
 	return 0;
@@ -686,26 +854,48 @@ int ioq_closedir(struct ioq *ioq, struct bfs_dir *dir, void *ptr) {
 	return 0;
 }
 
-struct ioq_ent *ioq_pop(struct ioq *ioq) {
-	if (ioq->size == 0) {
-		return NULL;
+int ioq_stat(struct ioq *ioq, int dfd, const char *path, enum bfs_stat_flags flags, struct bfs_stat *buf, void *ptr) {
+	struct ioq_ent *ent = ioq_request(ioq, IOQ_STAT, ptr);
+	if (!ent) {
+		return -1;
 	}
 
-	return ioqq_pop(ioq->ready);
+	struct ioq_stat *args = &ent->stat;
+	args->dfd = dfd;
+	args->path = path;
+	args->flags = flags;
+	args->buf = buf;
+
+#if BFS_USE_LIBURING && BFS_USE_STATX
+	args->xbuf = arena_alloc(&ioq->xbufs);
+	if (!args->xbuf) {
+		ioq_free(ioq, ent);
+		return -1;
+	}
+#endif
+
+	ioqq_push(ioq->pending, ent);
+	return 0;
 }
 
-struct ioq_ent *ioq_trypop(struct ioq *ioq) {
+struct ioq_ent *ioq_pop(struct ioq *ioq, bool block) {
 	if (ioq->size == 0) {
 		return NULL;
 	}
 
-	return ioqq_trypop(ioq->ready);
+	return ioqq_pop(ioq->ready, block);
 }
 
 void ioq_free(struct ioq *ioq, struct ioq_ent *ent) {
 	bfs_assert(ioq->size > 0);
 	--ioq->size;
 
+#if BFS_USE_LIBURING && BFS_USE_STATX
+	if (ent->op == IOQ_STAT) {
+		arena_free(&ioq->xbufs, ent->stat.xbuf);
+	}
+#endif
+
 	arena_free(&ioq->ents, ent);
 }
 
@@ -735,6 +925,9 @@ void ioq_destroy(struct ioq *ioq) {
 	ioqq_destroy(ioq->ready);
 	ioqq_destroy(ioq->pending);
 
+#if BFS_USE_LIBURING && BFS_USE_STATX
+	arena_destroy(&ioq->xbufs);
+#endif
 	arena_destroy(&ioq->ents);
 
 	free(ioq);
