@@ -8,6 +8,9 @@
 #ifndef BFS_IOQ_H
 #define BFS_IOQ_H
 
+#include "config.h"
+#include "dir.h"
+#include "stat.h"
 #include <stddef.h>
 
 /**
@@ -25,6 +28,8 @@ enum ioq_op {
 	IOQ_OPENDIR,
 	/** ioq_closedir(). */
 	IOQ_CLOSEDIR,
+	/** ioq_stat(). */
+	IOQ_STAT,
 };
 
 /**
@@ -34,10 +39,8 @@ struct ioq_ent {
 	/** The I/O operation. */
 	enum ioq_op op;
 
-	/** The return value of the operation. */
-	int ret;
-	/** The error code, if the operation failed. */
-	int error;
+	/** The return value (on success) or negative error code (on failure). */
+	int result;
 
 	/** Arbitrary user data. */
 	void *ptr;
@@ -53,11 +56,20 @@ struct ioq_ent {
 			struct bfs_dir *dir;
 			int dfd;
 			const char *path;
+			enum bfs_dir_flags flags;
 		} opendir;
 		/** ioq_closedir() args. */
 		struct ioq_closedir {
 			struct bfs_dir *dir;
 		} closedir;
+		/** ioq_stat() args. */
+		struct ioq_stat {
+			int dfd;
+			const char *path;
+			enum bfs_stat_flags flags;
+			struct bfs_stat *buf;
+			void *xbuf;
+		} stat;
 	};
 };
 
@@ -103,12 +115,14 @@ int ioq_close(struct ioq *ioq, int fd, void *ptr);
  *         The base file descriptor.
  * @param path
  *         The path to open, relative to dfd.
+ * @param flags
+ *         Flags that control which directory entries are listed.
  * @param ptr
  *         An arbitrary pointer to associate with the request.
  * @return
  *         0 on success, or -1 on failure.
  */
-int ioq_opendir(struct ioq *ioq, struct bfs_dir *dir, int dfd, const char *path, void *ptr);
+int ioq_opendir(struct ioq *ioq, struct bfs_dir *dir, int dfd, const char *path, enum bfs_dir_flags flags, void *ptr);
 
 /**
  * Asynchronous bfs_closedir().
@@ -124,6 +138,26 @@ int ioq_opendir(struct ioq *ioq, struct bfs_dir *dir, int dfd, const char *path,
  */
 int ioq_closedir(struct ioq *ioq, struct bfs_dir *dir, void *ptr);
 
+/**
+ * Asynchronous bfs_stat().
+ *
+ * @param ioq
+ *         The I/O queue.
+ * @param dfd
+ *         The base file descriptor.
+ * @param path
+ *         The path to stat, relative to dfd.
+ * @param flags
+ *         Flags that affect the lookup.
+ * @param buf
+ *         A place to store the stat buffer, if successful.
+ * @param ptr
+ *         An arbitrary pointer to associate with the request.
+ * @return
+ *         0 on success, or -1 on failure.
+ */
+int ioq_stat(struct ioq *ioq, int dfd, const char *path, enum bfs_stat_flags flags, struct bfs_stat *buf, void *ptr);
+
 /**
  * Pop a response from the queue.
  *
@@ -132,17 +166,7 @@ int ioq_closedir(struct ioq *ioq, struct bfs_dir *dir, void *ptr);
  * @return
  *         The next response, or NULL.
  */
-struct ioq_ent *ioq_pop(struct ioq *ioq);
-
-/**
- * Pop a response from the queue, without blocking.
- *
- * @param ioq
- *         The I/O queue.
- * @return
- *         The next response, or NULL.
- */
-struct ioq_ent *ioq_trypop(struct ioq *ioq);
+struct ioq_ent *ioq_pop(struct ioq *ioq, bool block);
 
 /**
  * Free a queue entry.
