@@ -1,2 +1,2 @@
 # Make sure -xtype is considered side-effecting for facts_when_impure
-fail invoke_bfs loops -xtype l -depth 100
+! invoke_bfs loops -xtype l -depth 100
